"""Abstract matcher for CVEs."""

import logging
from datetime import date
from typing import Dict, Optional

from packaging import version

import abstracts

from aio.core.functional import utils

from aio.api.nist import abstract, typing


logger = logging.getLogger(__name__)


class ACVEMatcher(metaclass=abstracts.Abstraction):
    """Matcher for CVEs against date and version."""

    def __init__(
            self,
            filter_dict: typing.TrackedCPEMatchingFilterDict) -> None:
        self._filter_dict = filter_dict

    def __call__(
            self,
            cve: "abstract.ACVE",
            cpe: "abstract.ACPE",
            cpe_match: "typing.CVENodeMatchDict") -> bool:
        matched = self.match_cpe(cve, cpe, cpe_match)
        logger.debug(
            f"Matched\n  {self._match_debug(cve, cpe, cpe_match)}"
            if matched
            else f"No match\n  {self._match_debug(cve, cpe, cpe_match)}")
        return matched

    def __str__(self) -> str:
        return "/".join([
            self._truncate_cpe(self.tracked_cpe),
            str(self.tracked_date),
            str(self.tracked_version)])

    @property
    def tracked_cpe(self) -> "abstract.ACPE":
        """Tracked CPE object to match against."""
        return self._filter_dict["cpe"]

    @property
    def tracked_date(self) -> Optional[date]:
        """Publication date of tracked CPE (optional)."""
        return self._filter_dict["date"]

    @property
    def tracked_version(self) -> Optional[version.Version]:
        """Version of tracked CPE (optional)."""
        return self._filter_dict.get("version")

    def get_version_info(
            self,
            cpe_match: "typing.CVENodeMatchDict") -> Dict[
                str,
                Optional[version.Version]]:
        """Dictionary of mangled start/end versions for a CVE."""
        return {
            f"{action}_{ending}": self._cpe_version(cpe_match, action, ending)
            for action in ["end", "start"]
            for ending in ["exc", "inc"]}

    def match_cpe(
            self,
            cve: "abstract.ACVE",
            cpe: "abstract.ACPE",
            cpe_match: "typing.CVENodeMatchDict") -> bool:
        return bool(
            self.match_date(cve, cpe, cpe_match)
            and self.match_parts(cve, cpe, cpe_match)
            and self.match_version(cve, cpe, cpe_match))

    def match_date(
            self,
            cve: "abstract.ACVE",
            cpe: "abstract.ACPE",
            cpe_match: "typing.CVENodeMatchDict") -> bool:
        """Check whether date for provided CVE/CPE data matches the tracked
        date."""
        return not (
            self.tracked_date
            and (cpe.version == "*"
                 and (cve.published_date
                      < self.tracked_date)))

    def match_parts(
            self,
            cve: "abstract.ACVE",
            cpe: "abstract.ACPE",
            cpe_match: "typing.CVENodeMatchDict") -> bool:
        return (
            cpe.part == self.tracked_cpe.part
            and cpe.vendor == self.tracked_cpe.vendor
            and (
                self.tracked_cpe.product == "*"
                or cpe.product == self.tracked_cpe.product))

    def match_version(
            self,
            cve: "abstract.ACVE",
            cpe: "abstract.ACPE",
            cpe_match: "typing.CVENodeMatchDict") -> bool:
        """Check whether version for provided CVE/CPE data matches the tracked
        version."""
        if not self.tracked_version:
            return True
        version_info = self.get_version_info(cpe_match)
        return not (
            (version_info["end_exc"] is not None
             and self.tracked_version >= version_info["end_exc"])
            or (version_info["end_inc"] is not None
                and self.tracked_version > version_info["end_inc"])
            or (version_info["start_exc"] is not None
                and self.tracked_version <= version_info["start_exc"])
            or (version_info["start_inc"] is not None
                and self.tracked_version < version_info["start_inc"]))

    def _cpe_version(
            self,
            cpe_match: "typing.CVENodeMatchDict",
            action: str,
            ending: str) -> Optional[version.Version]:
        """Transform a cpe version match key -> semantic version."""
        version_info = cpe_match.get(
            f"version{action.capitalize()}{ending.capitalize()}luding", None)
        return (
            version.Version(utils.typed(str, version_info))
            if version_info is not None
            else None)

    def _match_debug(
            self,
            cve: "abstract.ACVE",
            cpe: "abstract.ACPE",
            cpe_match: "typing.CVENodeMatchDict") -> str:
        match_info = "/".join([
            cve.id,
            self._truncate_cpe(cpe),
            str(cve.published_date),
            str(cpe_match)])
        return f"{self}\n  -> {match_info}"

    def _truncate_cpe(
            self,
            cpe: "abstract.ACPE") -> str:
        return str(cpe).split(":", 2)[2]




from datetime import date
from functools import cached_property
from typing import Dict, List, Optional, Set, Type

import abstracts

from aio.api.nist import abstract, typing
from aio.core.functional import QueryDict


class ACVE(metaclass=abstracts.Abstraction):
    """A CVE entry that matches against provided CPEs."""

    @classmethod
    def parse_date(cls, date_str: str) -> date:
        return date.fromisoformat(date_str.split('T')[0])

    def __init__(
            self,
            cve_data: Dict,
            tracked_cpes: "typing.TrackedCPEMatchingDict",
            cpe_class: Type["abstract.ACPE"],
            cve_fields: Optional[QueryDict] = None) -> None:
        self.cve_data = cve_data
        self.tracked_cpes = tracked_cpes
        self.cpe_class = cpe_class
        self.cve_fields = cve_fields

    @cached_property
    def cpes(self) -> Set["abstract.ACPE"]:
        """Associated CPEs."""
        cpe_set: Set["abstract.ACPE"] = set()
        self.gather_cpes(self.nodes, cpe_set)
        return cpe_set

    @cached_property
    def cve_dict(self) -> Dict:
        """Dictionary of parsed information about this CVE."""
        return dict(
            **self.cve_data,
            cpes=self.gathered_cpes)

    @property
    def gathered_cpes(self) -> "typing.CPEsTuple":
        """Tuple of matching CVEs for this CVE."""
        return tuple(
            dict(part=cpe.part,
                 vendor=cpe.vendor,
                 product=cpe.product,
                 version=cpe.version)
            for cpe in self.cpes)

    @property
    def id(self) -> str:
        """CVE ID."""
        return self.cve_data["id"]

    @property
    def nodes(self) -> List["typing.CVENodeDict"]:
        """CVE nodes used for matching versions."""
        return self.cve_data["nodes"]

    @property
    def published_date(self) -> date:
        """Published date of this CVE."""
        return self.parse_date(self.cve_data["published_date"])

    def gather_cpes(
            self,
            nodes: List["typing.CVENodeDict"],
            cpe_set: Set["abstract.ACPE"]) -> None:
        """Recursively gather CPE data from CVE nodes."""
        for node in nodes:
            for cpe_match in node.get('cpe_match', []):
                cpe = self.cpe_class.from_string(cpe_match.pop('cpe23Uri'))
                if self.include_version(cpe_match, cpe):
                    cpe_set.add(cpe)
            children = node.get('children', [])
            if children:
                self.gather_cpes(children, cpe_set)

    def include_version(
            self,
            cpe_match: "typing.CVENodeMatchDict",
            cpe: "abstract.ACPE") -> bool:
        """Determine whether a CPE matches according to installed version of a
        dependency."""
        return (
            self.tracked_cpes[str(cpe)](self, cpe, cpe_match)
            if str(cpe) in self.tracked_cpes
            else False)

    def update_fields(self, data) -> "typing.CVEDict":
        """Add requested fields for matched CVE."""
        # todo: if nodes are requested dont delete just rename if required
        del self.cve_dict["nodes"]
        if self.cve_fields:
            self.cve_dict.update(**self.cve_fields(data))



"""Abstract CPE."""

from functools import cached_property, lru_cache

import abstracts

from aio.api.nist import exceptions


class ACPE(metaclass=abstracts.Abstraction):
    """Model a subset of CPE fields that are used in CPE matching."""

    @classmethod
    @lru_cache(maxsize=None)
    def from_string(cls, cpe_str: str) -> "ACPE":
        """Generate a CPE object from a CPE string."""
        components = cpe_str.split(':')
        if len(components) < 6 or not cpe_str.startswith('cpe:2.3:'):
            raise exceptions.CPEError(
                f"CPE string ({cpe_str}) must be a valid CPE v2.3 string")
        return cls(*components[2:6])

    def __init__(
            self,
            part: str,
            vendor: str,
            product: str = "*",
            version: str = "*") -> None:
        self.part = part
        self.vendor = vendor
        self.product = product
        self.version = version

    def __str__(self) -> str:
        return (
            f"cpe:2.3:{self.part}:{self.vendor}:"
            f"{self.product}:{self.version}")

    @cached_property
    def vendor_normalized(self) -> str:
        """Return a normalized CPE where only part and vendor are
        significant."""
        return str(self.__class__(self.part, self.vendor, '*', '*'))



"""NIST CVE data parser."""

import logging
from datetime import date
from functools import cached_property
from typing import Dict, Iterator, List, Optional, Set, Tuple, Type

import abstracts

from aio.api.nist import abstract, typing
from aio.core.functional import utils, qdict, QueryDict


logger = logging.getLogger(__name__)


class ANISTParser(metaclass=abstracts.Abstraction):
    """Sync blocking CVE data parser.

    Run me in an executor!.
    """

    def __init__(
            self,
            tracked_cpes: "typing.TrackedCPEDict",
            cve_fields: Optional[QueryDict] = None,
            ignored_cves: Optional[Set] = None) -> None:
        self._tracked_cpes = tracked_cpes
        self.ignored_cves = ignored_cves or set()
        self.cve_fields = cve_fields

    def __call__(
            self,
            data: bytes) -> "typing.CVEDataTuple":
        """Slow, blocking parser."""
        return self.parse_cve_data(data)

    @property  # type:ignore
    @abstracts.interfacemethod
    def cpe_class(self) -> "abstract.ACPE":
        """CPE class."""
        raise NotImplementedError

    @cached_property
    def cpe_revmap(self) -> "typing.CPERevmapDict":
        """Collected reverse mapping of CPEs."""
        return {}

    @property  # type:ignore
    @abstracts.interfacemethod
    def cve_class(self) -> "abstract.ACVE":
        """CVE class."""
        raise NotImplementedError

    @cached_property
    def cves(self) -> "typing.CVEDict":
        """Collected CVEs."""
        return {}

    @property  # type:ignore
    @abstracts.interfacemethod
    def matcher_class(self) -> Type["abstract.ACVEMatcher"]:
        """Version matcher class."""
        raise NotImplementedError

    @cached_property
    def query_fields(self) -> "QueryDict":
        """Fields used for initial matching of the CVE."""
        return qdict(
            id="cve/CVE_data_meta/ID",
            nodes="configurations/nodes",
            published_date="publishedDate")

    @cached_property
    def tracked_cpes(self) -> "typing.TrackedCPEMatchingDict":
        """Mapping of tracked CPEs, with a filter dictionary for matching."""
        return {
            k: self._tracked_cpe(k, v)
            for k, v
            in self._tracked_cpes.items()}

    def add_cpe_revmap(self, cve: "abstract.ACVE") -> None:
        """Add a CVE item to the CPE reverse mapping."""
        for cve_cpe in cve.cpes:
            self.cpe_revmap[cve_cpe.vendor_normalized] = self.cpe_revmap.get(
                cve_cpe.vendor_normalized, set())
            self.cpe_revmap[cve_cpe.vendor_normalized].add(cve.id)

    def add_cve(
            self,
            cve_item: "typing.CVEItemDict",
            cve_data: Dict) -> None:
        """Add CVE to the the cve list if it has relevant CPEs."""
        cve = self.cve_class(
            cve_data,
            self.tracked_cpes,
            self.cpe_class,
            self.cve_fields)
        if not cve.cpes:
            return
        cve.update_fields(cve_item)
        self.add_cpe_revmap(cve)
        self.cves[cve.id] = cve.cve_dict

    def include_cve(self, cve: "typing.CVEItemDict") -> bool:
        """Determine whether the CVE should be included for further
        analysis."""
        if not cve["configurations"]["nodes"]:
            # This will not have any CPEs - not sure what type of record it is
            return False
        is_v3 = bool(cve['impact'].get('baseMetricV3'))
        id = cve['cve']['CVE_data_meta']['ID']
        if is_v3 and id not in self.ignored_cves:
            return True
        logger.debug(
            f"Excluding {id} "
            f"(v3: {is_v3})")
        return False

    def iter_cve_json(
            self,
            data: bytes) -> Iterator[
                Tuple["typing.CVEItemDict",
                      "typing.CVEQueryDict"]]:
        """List of CVE dicts with information required to parse CPEs and
        filter."""
        for cve_item in self._junzip_items(data):
            if self.include_cve(cve_item):
                cve_data = self.query_fields(cve_item)
                # This is too verbose, but can be helpful for debugging
                # logger.debug(f"Analyze CVE {parsed['id']}")
                yield cve_item, cve_data

    def parse_cve_data(
            self,
            data: bytes) -> "typing.CVEDataTuple":
        """Parse CVE JSON dictionary."""
        for cve_item, cve_data in self.iter_cve_json(data):
            self.add_cve(cve_item, cve_data)
        return self.cves, self.cpe_revmap

    def _iso_date(self, date_str: Optional[str]) -> Optional[date]:
        return (
            self.cve_class.parse_date(date_str)
            if date_str
            else None)

    def _junzip_items(self, data: bytes) -> List["typing.CVEItemDict"]:
        return utils.junzip(data)["CVE_Items"]

    def _tracked_cpe(
            self,
            cpe_name: str,
            cpe_filter: typing.TrackedCPEFilterDict) -> (
                "abstract.ACVEMatcher"):
        match_dict: typing.TrackedCPEMatchingFilterDict = dict(
            version=cpe_filter.get("version"),
            cpe=self.cpe_class.from_string(cpe_name),
            date=self._iso_date(cpe_filter.get("date")))
        return self.matcher_class(match_dict)



"""NIST CVE data downloader."""

import asyncio
import logging
from concurrent import futures
from datetime import datetime
from functools import cached_property
from typing import AsyncIterator, Optional, Set, Type

import aiohttp

import abstracts

from aio.core import event
from aio.core.functional import async_property, QueryDict
from aio.core.tasks import concurrent

from aio.api.nist import abstract, typing


logger = logging.getLogger(__name__)

NIST_URL_TPL = (
    "https://nvd.nist.gov/feeds/json/cve/1.1/nvdcve-1.1-{year}.json.gz")
SCAN_FROM_YEAR = 2018


@abstracts.implementer(event.IReactive)
class ANISTDownloader(event.AReactive, metaclass=abstracts.Abstraction):

    def __init__(
            self,
            tracked_cpes: "typing.TrackedCPEDict",
            cve_fields: Optional[QueryDict] = None,
            ignored_cves: Optional[Set] = None,
            since: Optional[int] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            pool: Optional[futures.Executor] = None,
            session: Optional[aiohttp.ClientSession] = None) -> None:
        self._since = since
        self.cve_fields = cve_fields
        self.ignored_cves = ignored_cves
        self.tracked_cpes = tracked_cpes
        self._session = session
        self._loop = loop
        self._pool = pool

    async def __aiter__(self) -> AsyncIterator[str]:
        async for url in self.downloads:
            yield url

    @cached_property
    def cpe_revmap(self) -> "typing.CPERevmapDict":
        """Collected reverse mapping of CPEs."""
        return {}

    @cached_property
    def cves(self) -> "typing.CVEDict":
        """Collected CVEs."""
        return {}

    @property
    def downloaders(self) -> "typing.DownloadGenerator":
        """Download co-routines for NIST data."""
        for url in self.urls:
            yield self.download_and_parse(url)

    @async_property
    async def downloads(self) -> AsyncIterator[str]:
        """CVE data derived from parsing NIST CVE data."""
        async for download in concurrent(self.downloaders):
            yield download.url

    @property
    def nist_url_tpl(self) -> str:
        """Default URL template string for NIST downloads."""
        return NIST_URL_TPL

    @cached_property
    def parser(self) -> "abstract.ANISTParser":
        """NIST CVE parser, to be run in processor pool."""
        return self.parser_class(
            self.tracked_cpes,
            cve_fields=self.cve_fields,
            ignored_cves=self.ignored_cves)

    @property  # type:ignore
    @abstracts.interfacemethod
    def parser_class(self) -> Type["abstract.ANISTParser"]:
        """NIST parser class."""
        raise NotImplementedError

    @property
    def scan_year_end(self) -> int:
        """Inclusive end year to scan to."""
        return datetime.now().year + 1

    @cached_property
    def session(self) -> aiohttp.ClientSession:
        """HTTP client session."""
        return self._session or aiohttp.ClientSession()

    @property
    def since(self) -> int:
        return max(
            self._since or SCAN_FROM_YEAR,
            SCAN_FROM_YEAR)

    @property
    def urls(self) -> Set[str]:
        """URLs to fetch NIST data from."""
        return set(
            self.nist_url_tpl.format(year=year)
            for year
            in self.years)

    @property
    def years(self) -> range:
        """Range of years to scan."""
        return range(
            self.since,
            self.scan_year_end)

    def add(
            self,
            cves: "typing.CVEDict",
            cpe_revmap: "typing.CPERevmapDict") -> None:
        """Capture incoming CVE data."""
        self.cves.update(cves)
        self.cpe_revmap.update(cpe_revmap)

    async def download_and_parse(self, url: str) -> aiohttp.ClientResponse:
        """Async download and parsing of CVE data."""
        download = await self.session.get(url)
        logger.debug(f"Downloading CVE data: {url}")
        self.add(*await self.parse(url, await download.read()))
        logger.debug(f"CVE data saved: {url}")
        return download

    async def parse(self, url: str, data: bytes) -> "typing.CVEDataTuple":
        """Parse incoming data in executor."""
        # Disable this comment to prevent running the parser in a separate
        # process - useful for debugging.
        # return self.parser(data)
        logger.debug(f"Parsing CVE data: {url}")
        return await self.loop.run_in_executor(
            self.pool,
            self.parser,
            data)




from typing import Type

import abstracts

from aio.api import nist


@abstracts.implementer(nist.ACPE)
class CPE:
    pass


@abstracts.implementer(nist.ACVE)
class CVE:
    pass


@abstracts.implementer(nist.ACVEMatcher)
class CVEMatcher:
    pass


@abstracts.implementer(nist.ANISTDownloader)
class NISTDownloader:

    @property
    def parser_class(self) -> Type[nist.ANISTParser]:
        return NISTParser


@abstracts.implementer(nist.ANISTParser)
class NISTParser:

    @property
    def cpe_class(self) -> Type[nist.ACPE]:
        return CPE

    @property
    def cve_class(self) -> Type[nist.ACVE]:
        return CVE

    @property
    def matcher_class(self) -> Type[nist.ACVEMatcher]:
        return CVEMatcher



"""aio.api.nist."""

from . import abstract, exceptions, typing
from .abstract import (
    ACPE,
    ACVE,
    ACVEMatcher,
    ANISTDownloader,
    ANISTParser)
from .nist import (
    CPE,
    CVE,
    CVEMatcher,
    NISTDownloader,
    NISTParser)

__all__ = (
    "ACPE",
    "ACVE",
    "ACVEMatcher",
    "ANISTDownloader",
    "ANISTParser",
    "abstract",
    "CPE",
    "CVE",
    "CVEMatcher",
    "NISTDownloader",
    "NISTParser",
    "exceptions",
    "typing")



from datetime import date
from typing import (
    Any, Coroutine, Dict, Generator, List, Optional, Set,
    Tuple, TypedDict)

from packaging import version

import aiohttp

from aio.api.nist import abstract, typing


# NDIST CVE data format

class CVENodeMatchDict(TypedDict, total=False):
    cpe23Uri: str
    versionStartIncluding: str
    versionEndIncluding: str
    versionStartExcluding: str
    versionEndExcluding: str


class CVENodeDict(TypedDict, total=False):
    cpe_match: List[CVENodeMatchDict]
    children: List["typing.CVENodeDict"]  # type:ignore


class CVEItemConfigurationsDict(TypedDict, total=False):
    nodes: List[CVENodeDict]


class CVEItemDict(TypedDict, total=False):
    id: str
    configurations: CVEItemConfigurationsDict
    cve: Dict
    impact: Dict


class CVEJsonDict(TypedDict, total=False):
    CVE_Items: List[CVEItemDict]


class TrackedCPEFilterDict(TypedDict, total=False):
    version: Optional[version.Version]
    date: Optional[str]


class TrackedCPEMatchingFilterDict(TypedDict, total=False):
    version: Optional[version.Version]
    date: Optional[date]
    cpe: "abstract.ACPE"


class CPEDict(TypedDict):
    part: str
    product: str
    vendor: str
    version: str


CPEsTuple = Tuple[CPEDict, ...]
CPERevmapDict = Dict[str, Set[str]]
CVEDict = Dict[str, Dict]
CVEQueryDict = Dict[str, Dict]
CVEDataTuple = Tuple[CVEDict, CPERevmapDict]
TrackedCPEDict = Dict[str, TrackedCPEFilterDict]
TrackedCPEMatchingDict = Dict[str, "abstract.ACVEMatcher"]
DownloadGenerator = Generator[
    Coroutine[Any, Any, aiohttp.ClientResponse],
    str,
    None]




from .cpe import ACPE
from .cve import ACVE
from .downloader import ANISTDownloader
from .matcher import ACVEMatcher
from .parser import ANISTParser


__all__ = (
    "ACPE",
    "ACVE",
    "ACVEMatcher",
    "ANISTDownloader",
    "ANISTParser")
